@@ -1,6 +1,7 @@
+from __future__ import annotations
 from contextlib import _GeneratorContextManager
 from sqlite3 import register_adapter
-from typing import Any, Optional, Self, Literal
+from typing import Any, Iterator, Optional, Self, Literal, overload
 import datetime as _datetime
 import numpy as np
 from pendulum import (
@@ -96,6 +97,125 @@ class Timezone(BaseTimezone):
 UTC = Timezone(key="UTC")
 
 
+class Date(BaseDate):
+    @classmethod
+    def from_base(
+        cls,
+        base: BaseDate,
+    ) -> Self:
+        return cls(
+            year=base.year,
+            month=base.month,
+            day=base.day,
+        )
+
+    @classmethod
+    def from_datetime(
+        cls,
+        date: _datetime.date,
+    ) -> Self:
+        return cls(
+            year=date.year,
+            month=date.month,
+            day=date.day,
+        )
+
+    @classmethod
+    def parse(
+        cls,
+        input,
+    ) -> Self:
+        output = parse(input=input)
+
+        if not isinstance(output, cls):
+            raise ValueError("Could not parse to date")
+
+        return output
+
+    def to_datetime(
+        self,
+        tz: str | Timezone = UTC,
+    ) -> DateTime:
+        return DateTime.create(
+            year=self.year,
+            month=self.month,
+            day=self.day,
+            hour=0,
+            minute=0,
+            second=0,
+            microsecond=0,
+            tz=tz,
+        )
+
+    def to_numpy(
+        self,
+    ) -> np.datetime64:
+        return np.datetime64(self)
+
+
+class Time(BaseTime):
+    @classmethod
+    def from_base(
+        cls,
+        base: BaseTime,
+    ) -> Self:
+        return cls(
+            hour=base.hour,
+            minute=base.minute,
+            second=base.second,
+            microsecond=base.microsecond,
+            tzinfo=base.tzinfo,
+        )
+
+    @classmethod
+    def from_datetime(
+        cls,
+        time: _datetime.time,
+        tz: str | Timezone | FixedTimezone | _datetime.tzinfo | None = UTC,
+    ) -> Self:
+        return cls.instance(
+            t=time,
+            tz=tz,
+        )
+
+    @classmethod
+    def parse(
+        cls,
+        input,
+    ) -> Self:
+        output = parse(input=input)
+
+        if not isinstance(output, cls):
+            raise ValueError("Could not parse to time")
+
+        return output
+
+    def today(
+        self,
+    ) -> DateTime:
+        return DateTime.now(tz=self.tzinfo).at(
+            hour=self.hour,
+            minute=self.minute,
+            second=self.second,
+            microsecond=self.microsecond,
+        )
+
+    def on(
+        self,
+        date: Date,
+    ) -> DateTime:
+        return DateTime(
+            year=date.year,
+            month=date.month,
+            day=date.day,
+            hour=self.hour,
+            minute=self.minute,
+            second=self.second,
+            microsecond=self.microsecond,
+            tzinfo=self.tzinfo,
+        )
+
+
 class DateTime(BaseDateTime):
     @classmethod
     def from_base(
@@ -277,131 +397,27 @@ class DateTime(BaseDateTime):
             microsecond=self.microsecond,
         )
 
-    def numpy(
-        self,
-    ) -> np.datetime64:
-        return np.datetime64(self)
-
-
-class Date(BaseDate):
-    @classmethod
-    def from_base(
-        cls,
-        base: BaseDate,
-    ) -> Self:
-        return cls(
-            year=base.year,
-            month=base.month,
-            day=base.day,
-        )
-
-    @classmethod
-    def from_datetime(
-        cls,
-        date: _datetime.date,
-    ) -> Self:
-        return cls(
-            year=date.year,
-            month=date.month,
-            day=date.day,
-        )
-
-    @classmethod
-    def parse(
-        cls,
-        input,
-    ) -> Self:
-        output = parse(input=input)
-
-        if not isinstance(output, cls):
-            raise ValueError("Could not parse to date")
-
-        return output
-
-    def to_datetime(
-        self,
-        tz: str | Timezone = UTC,
-    ) -> DateTime:
-        return DateTime.create(
+    def date(self) -> Date:
+        return Date(
             year=self.year,
             month=self.month,
             day=self.day,
-            hour=0,
-            minute=0,
-            second=0,
-            microsecond=0,
-            tz=tz,
         )
 
-    def numpy(
+    def time(self) -> Time:
+        return Time(
+            hour=self.hour,
+            minute=self.minute,
+            second=self.second,
+            microsecond=self.microsecond,
+        )
+
+    def to_numpy(
         self,
     ) -> np.datetime64:
         return np.datetime64(self)
 
 
-class Time(BaseTime):
-    @classmethod
-    def from_base(
-        cls,
-        base: BaseTime,
-    ) -> Self:
-        return cls(
-            hour=base.hour,
-            minute=base.minute,
-            second=base.second,
-            microsecond=base.microsecond,
-            tzinfo=base.tzinfo,
-        )
-
-    @classmethod
-    def from_datetime(
-        cls,
-        time: _datetime.time,
-        tz: str | Timezone | FixedTimezone | _datetime.tzinfo | None = UTC,
-    ) -> Self:
-        return cls.instance(
-            t=time,
-            tz=tz,
-        )
-
-    @classmethod
-    def parse(
-        cls,
-        input,
-    ) -> Self:
-        output = parse(input=input)
-
-        if not isinstance(output, cls):
-            raise ValueError("Could not parse to time")
-
-        return output
-
-    def today(
-        self,
-    ) -> DateTime:
-        return DateTime.now(tz=self.tzinfo).at(
-            hour=self.hour,
-            minute=self.minute,
-            second=self.second,
-            microsecond=self.microsecond,
-        )
-
-    def on(
-        self,
-        date: Date,
-    ) -> DateTime:
-        return DateTime(
-            year=date.year,
-            month=date.month,
-            day=date.day,
-            hour=self.hour,
-            minute=self.minute,
-            second=self.second,
-            microsecond=self.microsecond,
-            tzinfo=self.tzinfo,
-        )
-
-
 # TODO: NewInterval for working with pandas? nah add to pandas accessor? Or maybe to convert to pandas indexable thing
 class Interval(BaseInterval):
     def __new__(
@@ -499,6 +515,46 @@ class Intervals(object):
 
         return self
 
+    def __repr__(
+        self,
+    ) -> str:
+        return f"Intervals(\n\t{'\n\t'.join([repr(interval) for interval in self.intervals])}\n)"
+
+    def __iter__(
+        self,
+    ) -> Iterator[Interval]:
+        return iter(self.intervals)
+
+    def __len__(
+        self,
+    ) -> int:
+        return len(self.intervals)
+
+    @overload
+    def __getitem__(
+        self,
+        key: int,
+    ) -> Interval: ...
+
+    @overload
+    def __getitem__(
+        self,
+        key: slice,
+    ) -> Intervals: ...
+
+    def __getitem__(
+        self,
+        key: int | slice,
+    ) -> Intervals | Interval:
+        if isinstance(key, slice):
+            return Intervals(*self.intervals[key])
+
+        elif isinstance(key, int):
+            return self.intervals[key]
+
+        else:
+            raise TypeError("Invalid key type")
+
 
 class Traveller(BaseTraveller):
     def __init__(
